/*
 MKMK: Makefile maker.
 By Luis Crespo, lcrespo@ergos.es
 Please read MKMK.TXT for a program description.

 Define "WATCOM_MAKE" for Watcom C/C++ oriented makefiles.
 Define "UNIX" for unix orieted makefiles
*/

#include <stdio.h>
#include <string.h>

#include "defs.h"
#include "contain.h"

#ifdef __WATCOMC__
  #include <dos.h>
  #define findfirst(x,y,z) _dos_findfirst(x,z,y)
  #define findnext _dos_findnext
  #define ffblk find_t
  #define ff_name name
#else
  #include <dir.h>
#endif

#define WATCOM_MAKE

#ifdef UNIX
  #define PathChar '/'
#else
  #define PathChar '\\'
#endif

#ifdef WATCOM_MAKE
  #define MoreChar '&'
#else
  #define MoreChar '\\'
#endif



//------------------------- Dependency list -------------------------

struct TDepItem {
  char Name[80];
  char Path[80];
};

class TDepList:public TDynArray {
  public:
    TDepList():TDynArray(20,10,sizeof(TDepItem)){}
    bool Search(char *s);
    void Insert(char *n, char *p);
    void AddMask(char *Mask);
};


void GetPath(char *s, char *p, int Len)
{
  int l;

  memset(p,0,Len);
  l=strlen(s)-1;
  if (l<1) return;
  while (l>=0 && s[l]!=PathChar) l--;
  if (l>=0 && l<Len) memcpy(p,s,l+1);
}

void FSplit(char *s, char *p, char *n)
{
  int l;

  l=strlen(s)-1;
  if (l<1) return;
  while (l>=0 && s[l]!=PathChar) l--;
  memcpy(p,s,l+1); p[l+1]=0;
  strcpy(n,&s[l+1]);
}


bool TDepList::Search(char *s)
{
  int i;
  bool Found;
  TDepItem *d;

  i=0;
  Found=false;
  while (i<Count && !Found)
  {
    d=(TDepItem*)(*this)[i];
    Found = strcmpi(d->Name,s)==0;
    i++;
  }
  return Found;
}

void TDepList::Insert(char *n, char *p)
{
  TDepItem i;

  memset(&i,0,sizeof(i));
  strncpy(i.Name,n,sizeof(i.Name)-1);
  strncpy(i.Path,p,sizeof(i.Path)-1);
  TDynArray::Insert(&i);
}

void TDepList::AddMask(char *Mask)
{
  TDepItem i;
  struct ffblk ffblk;
  bool Done;
  char Path[80],Name[80];

  memset(&i,0,sizeof(i));
  FSplit(Mask,Path,Name);
  strcpy(i.Name,Name);
  strcpy(i.Path,Path);
  if (strchr(Mask,'*') || strchr(Mask,'?'))
  {
    Done=findfirst(Mask,&ffblk,0);
    while (!Done)
    {
      strcpy(i.Name,ffblk.ff_name);
      TDynArray::Insert(&i);
      Done=findnext(&ffblk);
    }
  }
  else TDynArray::Insert(&i);
}



//------------------------- Include file parsing -------------------------

class TIncludeParser {
  public:
    bool Parse(char *FileName, bool CPP=true);
    TDepList l;
  private:
    char FilePath[80];
    bool Comment,bCPP;
    bool BasicParse(char *FileName);
    void ParseComments(char *s);
    void ParseCLine(char *s);
    void AddInclude(char *IncName);
};

void TIncludeParser::AddInclude(char *IncName)
{
  char s[160];
  bool Found;

  if (l.Search(IncName)) return;
  printf("    %s\n",IncName);
  Found=BasicParse(IncName);
  if (Found) l.Insert(IncName,"");
  else if (FilePath[0]!=0)
  {
    strcpy(s,FilePath);
    strcat(s,IncName);
    Found=BasicParse(s);
    l.Insert(IncName,FilePath);
  }
  if (!Found)
    fprintf(stderr,"WARNING: %s could not be opened\n",IncName);
}

void TIncludeParser::ParseCLine(char *s)
{
  char *t,*p1,*p2;

  t=strstr(s,"#include");
  if (t)
  {
    p1=strchr(t,'"');
    if (p1)
    {
      p1++;
      p2=strchr(p1,'"');
      if (p2)
      {
        *p2=0;
        AddInclude(p1);
      }
    }
  }
}

void TIncludeParser::ParseComments(char *s)
{
  int i;

  i=0;
  do {
    if (Comment)
    {
      if (s[i]=='*' && s[i+1]=='/')
      {
        Comment=false;
        i++;
      }
      else (s[i]='*');
    }
    else if (s[i]=='/')
    {
      switch (s[i+1])
      {
        case '*': Comment=true; i++; break;
        case '/': s[i]=0; i--; break;
      }
    }
    i++;
  } while (s[i]);
}

bool TIncludeParser::Parse(char *FileName, bool CPP)
{
  Comment=false;
  bCPP=CPP;
  GetPath(FileName,FilePath,sizeof(FilePath));
  if (!BasicParse(FileName))
  {
    fprintf(stderr,"WARNING: %s could not be opened\n",FileName);
    return false;
  }
  else return true;
}

bool TIncludeParser::BasicParse(char *FileName)
{
  FILE *f;
  char Line[256];

  f=fopen(FileName,"rt");
  if (!f) return false;
  do {
    fgets(Line,sizeof(Line)-1,f);
    if (Line[0]!=0)
    {
      ParseComments(Line);
      ParseCLine(Line);
    }
  } while (!feof(f));
  fclose(f);
  return true;
}




//------------------------------ Main code ------------------------------

void ShowHelp()
{
  puts("MKMK: makefile generator");
  puts("Usage: MKMK exename file1 [file2 file3...]\n");
}

void WriteHeader(FILE *f, char *ExeName)
{
  fprintf(f,"#Make file automatically generated by MKMK\n\n");
  fprintf(f,"main: %s\n\n",ExeName);
  fprintf(f,"clean:\n");
  fprintf(f,"  del *.obj\n");
  fprintf(f,"  del %s\n\n",ExeName);
  fprintf(f,"#Compiler and linker definitions (suggestion):\n");
#ifdef WATCOM_MAKE
  fprintf(f,"CC    = WPP386\n");
  fprintf(f,"C_OPS = /Oneatx /zp4 /5 /fp5 /D1 /D3 /W1 /MF /SG /HW /ZQ\n");
  fprintf(f,"LINK  = WLINK\n");
  fprintf(f,"L_OPS = SYSTEM DOS4G OPTION STACK=32768 DEBUG WATCOM A OPTION QUIET\n\n");
#else
  fprintf(f,"CC    = your favorite C/CPP compiler\n");
  fprintf(f,"C_OPS = your compiler options\n");
  fprintf(f,"LINK  = your favorite linker\n");
  fprintf(f,"L_OPS = your linker options\n\n");
#endif
  fprintf(f,"#File dependencies:\n\n");
}

void ObjExt(char *s, char *ObjName, int Len)
{
  int l;

  memset(ObjName,0,Len);
  l=strlen(s)-1;
  if (l<1) return;
  while (l>=0 && s[l]!='.') l--;
  if (l>=0 && l<Len) memcpy(ObjName,s,l+1);
  strcat(ObjName,"OBJ");
}

void ListFiles(FILE *f, char *Line, TDepList *l, bool IsObj=false)
{
  char s[120],FName[80];
  int i;
  TDepItem *d;

  for (i=0; i<l->Count; i++)
  {
    d=(TDepItem*)(*l)[i];
    strcpy(s," ");
    strcat(s,d->Path);
    if (IsObj) ObjExt(d->Name,FName,sizeof(FName));
    else strcpy(FName,d->Name);
    strcat(s,FName);
    if (strlen(s)+strlen(Line)>74)
    {
      fprintf(f,"%s %c\n",Line,MoreChar);
      sprintf(Line," %s",s);
    }
    else strcat(Line,s);
  }
  fprintf(f,"%s\n",Line);
}

void WriteDep(FILE *f, char *Name, TIncludeParser *p)
{
  char ObjName[80];
  char Line[120];

  ObjExt(Name,ObjName,sizeof(ObjName));
  sprintf(Line,"%s: %s",ObjName,Name);
  ListFiles(f,Line,&p->l);
  fprintf(f,"  $(CC) $(C_OPS) %s\n\n",Name);
}

void WriteEnding(FILE *f, char *ExeName, TDepList *l)
{
  char Line[120];

  fprintf(f,"OBJS =");
  ListFiles(f,Line,l,true);
  fprintf(f,"LIBS =\n\n");
  fprintf(f,"%s: $(OBJS) $(LIBS)\n",ExeName);
#ifdef WATCOM_MAKE
  fprintf(f,"  *$(LINK) $(L_OPS) name %s ",ExeName);
  fprintf(f,"file {$(OBJS)} library {$(LIBS)}\n");
#else
  fprintf(f,"  $(LINK) $(L_OPS) %s $(OBJS) $(LIBS)\n",ExeName);
#endif
}

void DoDep(FILE *f, TDepItem *d)
{
  TIncludeParser p;
  char FileName[256];

  strcpy(FileName,d->Path);
  strcat(FileName,d->Name);
  printf("\n%s\n",FileName);
  p.Parse(FileName);
  WriteDep(f,d->Name,&p);
}

void main(int argc, char *argv[])
{
  FILE *f;
  TDepList Sources;
  int i;

  if (argc<3)
  {
    ShowHelp();
    return;
  }
  f=fopen("makefile","wt");
  if (!f)
  {
	  fputs("ERROR: could not create makefile",stderr);
    return;
  }
  WriteHeader(f,argv[1]);
  for (i=2; i<argc; i++) Sources.AddMask(argv[i]);
  for (i=0; i<Sources.Count; i++) DoDep(f,(TDepItem*)Sources[i]);
  WriteEnding(f,argv[1],&Sources);
  fclose(f);
}
